/******************************************************************
 * Implementation of DES cryptography algorithm by Roman
 * Ver. 0.5
 * Date 03.12.2008
 ******************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "des.h"
#include "attack.h"

__inline__ uint64_t rdtsc(){
	uint32_t lo,hi;
	__asm__ __volatile__ 
		("rdtsc" : "=a" (lo), "=d" (hi));
	return (uint64_t)hi << 32 | lo;
}

void functional_test(void) {
	uint_fast64_t ks[16], k, p, c, r;

	k = 0x3030456789ABCDEFULL;
	p = 0xAA6F7720697320EEULL;
	c = 0x89081EB8E64C8951ULL;
	DES_keySchedule(0x3030456789ABCDEFULL, ks);
	r = DES_Encrypt(p, ks);
	if(r != c) {
		fprintf(stderr, "Error: DES(k=%016llX, p=%016llX) = %016llX (should be %016llX)\n", k, p, r, c);
		exit(-1);
	}
}

void CreateTimeDatabase(uint32_t nEnc, char* realtimeFilename, char* plaintextFilename)
{
	uint32_t iEnc = 0, i = 0;
	
	uint16_t nRep = 50;
	uint16_t iRep = 0;
	
	double a = 0, b = 0;
	double avg = 0.0;
	
	uint64_t plaintext = 0x0ULL, ciphertext = 0x0ULL;
	uint64_t ks[16];

	/*Extend the key*/
	DES_keySchedule(0x3030456789ABCDEFULL, ks);
	
	FILE *flPlaintext;
	FILE *flTime;
	
	flPlaintext = fopen(plaintextFilename,"w");
	if (flPlaintext == NULL){
		printf("Can't open the file %s.\n",realtimeFilename);
		exit(-1);
	}

	flTime 		= fopen(realtimeFilename, "w");
	if (flTime == NULL){
		printf("Can't open the file %s.\n",realtimeFilename);
		exit(-1);
	}

	/*Initialize RND*/
	srand ( rdtsc() );
	for (iEnc=0;iEnc<nEnc;iEnc++){
		/* generate random plaintext*/
		for (i=0;i<4;i++) {
			plaintext <<= 16;
			plaintext |= rand() % 65536;
		}
		
		/*Compute the average time that is take to perform encryption of the*/ 
		/*previously generated plaintext*/
		avg = 0.0;
		for(iRep = 0; iRep < nRep; iRep++) {
			a = rdtsc();
			ciphertext = DES_Encrypt(plaintext, ks);
			b = rdtsc();
			avg += (double)(b-a);
		}
		
		fprintf(flTime,"%f\n",avg/nRep);
		fprintf(flPlaintext,"%llX\n",plaintext);
		
		printf("Progress: %.2lf%%\r",(double)(iEnc+1)*(double)100/(double)(nEnc));
		fflush(stdout);	
	}
	
	fclose(flPlaintext);
	fclose(flTime);
}


int main(int argc, char **argv) {

	unsigned int nEnc = 100000000;

	char realtimeFilename[] 	= "../data/des_time.dat";
	char plaintextFilename[] 	= "../data/des_plaintext.dat";
	char resultFilename[] 		= "../data/pcc1.dat";
	uint8_t byteToRecover = 0;
	functional_test();
	
	/*CreateTimeDatabase(nEnc, realtimeFilename, plaintextFilename);*/

	TimingAttackDES_OneSbox(5000000, realtimeFilename, plaintextFilename, resultFilename, byteToRecover);

	return 0;
}
